//////////////////////////////////////////////////////////////////////////////
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//  XAPP 224 - Data Recovery
//
//////////////////////////////////////////////////////////////////////////////
//
//  File name :       data_recovery_virtex2_fast.v
//
//  Description :     Virtex2 data recovery module, uses two clocks 90 degrees apart to recover data
// 		      from incoming line data. (which is running slightly faster or slower than these
// 		      two clocks). More pipelined than basic version.
//
// 		      Output data bit 0 (sdata(0)) is valid when dv(0) is high
// 		      Output data bit 1 (sdata(1)) is valid when dv(1) is high
//
// 		      sample A is captured on : rising clk,    rising clk,   rising clk,   rising clk
// 		      sample B is captured on : rising clk90,  rising clk,   rising clk,   rising clk
// 		      sample C is captured on : falling clk,   rising clk90, rising clk,   rising clk
// 		      sample D is captured on : falling clk90, falling clk,  rising clk90, rising clk
//
//  Date - revision : May 16th 2003 - v 2.2
//
//  Author :          NJS
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778 
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
//              provided to you "as is". Xilinx and its licensors make and you
//              receive no warranties or conditions, express, implied,
//              statutory or otherwise, and Xilinx specifically disclaims any
//              implied warranties of merchantability, non-infringement,or
//              fitness for a particular purpose. Xilinx does not warrant that
//              the functions contained in these designs will meet your
//              requirements, or that the operation of these designs will be
//              uninterrupted or error free, or that defects in the Designs
//              will be corrected. Furthermore, Xilinx does not warrantor
//              make any representations regarding use or the results of the
//              use of the designs in terms of correctness, accuracy,
//              reliability, or otherwise.
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its
//              licensors be liable for any loss of data, lost profits,cost
//              or procurement of substitute goods or services, or for any
//              special, incidental, consequential, or indirect damages
//              arising from the use or operation of the designs or
//              accompanying documentation, however caused and on any theory
//              of liability. This limitation will apply even if Xilinx
//              has been advised of the possibility of such damage. This
//              limitation shall apply not-withstanding the failure of the
//              essential purpose of any limited remedies herein.
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved
//
//////////////////////////////////////////////////////////////////////////////
//
`timescale 1 ps / 1ps

module data_recovery_virtex2_fast (
	clk,
	clk90,
	data,
	rst,
	sdata,
	dv); 

input clk, clk90, data, rst ;
	
output 	[1 :0] 	sdata, dv ;
reg 	[1:0] 	sdata, dv ;

wire 		notclk ;
wire 		notclk90;
wire 	[3:0] 	az ;
wire 	[3:0] 	bz ;
wire 	[3:0] 	cz ;
wire 	[3:0] 	dz ;
reg 		az4 ;
reg 		bz4 ;
reg 		cz4 ;
reg 		dz4 ;
reg 		aap, aan ;
reg 		bbp, bbn ;
reg 		ccp, ccn ;
reg 		ddp, ddn ;
reg 		usea1, usea2, usea ;
reg 		useb1, useb2, useb ;
reg 		usec1, usec2, usec ;
reg 		used1, used2, used ;
reg 		useaa;
reg 		usebb;
reg 		usecc;
reg 		usedd;
reg 		useaaa;
reg 		usebbb;
reg 		useccc;
reg 		useddd;
reg 	[1:0] 	sdataa;
reg 	[1:0] 	sdatab ;
reg 	[1:0] 	sdatac;
reg 	[1:0] 	sdatad;
reg 	[1:0] 	dvalid;
reg 		pipe_ce0;
reg 		pipe_ce1;

assign notclk = ~clk ;
assign notclk90 = ~clk90 ;

always @(posedge clk or posedge rst)
begin
if (rst)
  begin
	dvalid <= 2'b00 ; dv <= 2'b00 ;
	az4 <= 1'b1 ; bz4 <= 1'b1 ; cz4 <= 1'b1 ; dz4 <= 1'b1 ;
	aap <= 1'b0  ; bbp <= 1'b0 ;  ccp <= 1'b0 ;  ddp <= 1'b0 ;	
	aan <= 1'b0  ; bbn <= 1'b0 ;  ccn <= 1'b0 ;  ddn <= 1'b0 ;	
	usea1 <= 1'b0  ; useb1 <= 1'b0  ; usec1 <= 1'b0  ; used1 <= 1'b0 ;
	usea2 <= 1'b0  ; useb2 <= 1'b0  ; usec2 <= 1'b0  ; used2 <= 1'b0 ;
	usea <= 1'b0  ; useb <= 1'b0  ; usec <= 1'b0  ; used <= 1'b0 ;
	useaa <= 1'b0 ; usebb <= 1'b0 ; usecc <= 1'b0 ; usedd <= 1'b0 ;
	useaaa <= 1'b0 ; usebbb <= 1'b0 ; useccc <= 1'b0 ; useddd <= 1'b0 ;
	sdataa <= 2'b11 ; sdatab <= 2'b11 ; sdatac <= 2'b11 ; sdatad <= 2'b11 ;
	sdata <= 2'b11 ; pipe_ce0 <= 1'b0 ; pipe_ce1 <= 1'b0 ;
  end // if
else begin
	az4 <= az[3] ;
	bz4 <= bz[3] ;
	cz4 <= cz[3] ;
	dz4 <= dz[3] ;
	aap <= (az[2] ^ az[3]) & ~az[2] ;	// find positive edges
	bbp <= (bz[2] ^ bz[3]) & ~bz[2] ;
	ccp <= (cz[2] ^ cz[3]) & ~cz[2] ;
	ddp <= (dz[2] ^ dz[3]) & ~dz[2] ;
	aan <= (az[2] ^ az[3]) & az[2] ;	// find negative edges
	bbn <= (bz[2] ^ bz[3]) & bz[2] ;
	ccn <= (cz[2] ^ cz[3]) & cz[2] ;
	ddn <= (dz[2] ^ dz[3]) & dz[2] ;
	usea1 <= bbp & ~ccp & ~ddp & aap ;
	useb1 <= ccp & ~ddp & aap & bbp ;
	usec1 <= ddp & aap & bbp & ccp ;
	used1 <= aap & ~bbp & ~ccp & ~ddp ;
	usea2 <= bbn & ~ccn & ~ddn & aan ;
	useb2 <= ccn & ~ddn & aan & bbn ;
	usec2 <= ddn & aan & bbn & ccn ;
	used2 <= aan & ~bbn & ~ccn & ~ddn ;
	usea <= usea1 | usea2 ;
	useb <= useb1 | useb2 ;
	usec <= usec1 | usec2 ;
	used <= used1 | used2 ;
	if (usea | useb | usec | used) begin
	   	pipe_ce0 <= 1'b1 ;
	   	pipe_ce1 <= 1'b1 ;
	end else begin
		pipe_ce1 <= 1'b0 ;
	end
	useaa <= usea ;
	usebb <= useb ;
	usecc <= usec ;
	usedd <= used ;
	if (pipe_ce1) begin
		useaaa <= useaa ;
		usebbb <= usebb ;
		useccc <= usecc ;
		useddd <= usedd ;
	end 
	sdataa[0] <= az4 & useaaa ; sdataa[1] <= az4 & useddd ;
	sdatab[0] <= bz4 & usebbb ; sdatab[1] <= 1'b0 ;
	sdatac[0] <= cz4 & useccc ; sdatac[1] <= 1'b0 ;
	sdatad[0] <= dz4 & useddd ; sdatad[1] <= dz4 & useaaa ;
	if (pipe_ce0) begin
		sdata <= sdataa | sdatab | sdatac | sdatad ;
	end
	if (useddd & useaa) begin
		dvalid <= 2'b00 ;
	end else if (useaaa & usedd) begin
		dvalid <= 2'b11 ; 
	end else begin
		dvalid <= 2'b01 ;
	end
	dv <= dvalid ;
  end // if
end //always

// get all the samples into the same time domain 

FDC ff_a0(.D(data),  .C(clk), .CLR(rst), .Q(az[0]))
/*synthesis rloc = "x0y0" iob = "FALSE" */ ; 
// synthesis attribute RLOC of ff_a0 "X0Y0"
// synthesis attribute IOB of ff_a0 "FALSE"
FDC ff_a1(.D(az[0]), .C(clk), .CLR(rst), .Q(az[1]))
/*synthesis rloc = "x0y0" */; // synthesis attribute RLOC of ff_a1 "X0Y0"
FDC ff_a2(.D(az[1]), .C(clk), .CLR(rst), .Q(az[2]))
/*synthesis rloc = "x3y1" */; // synthesis attribute RLOC of ff_a2 "X3Y1"
FDC ff_a3(.D(az[2]), .C(clk), .CLR(rst), .Q(az[3]))
/*synthesis rloc = "x3y1" */; // synthesis attribute RLOC of ff_a3 "X3Y1"

FDC ff_b0(.D(data),  .C(clk90), .CLR(rst), .Q(bz[0]))
/*synthesis rloc = "x1y0" iob = "FALSE" */ ; 
// synthesis attribute RLOC of ff_b0 "X1Y0"
// synthesis attribute IOB of ff_b0 "FALSE"
FDC ff_b1(.D(bz[0]), .C(clk),   .CLR(rst), .Q(bz[1]))
/*synthesis rloc = "x3y0" */; // synthesis attribute RLOC of ff_b1 "X3Y0"
FDC ff_b2(.D(bz[1]), .C(clk),   .CLR(rst), .Q(bz[2]))
/*synthesis rloc = "x3y0" */; // synthesis attribute RLOC of ff_b2 "X3Y0"
FDC ff_b3(.D(bz[2]), .C(clk),   .CLR(rst), .Q(bz[3]))
/*synthesis rloc = "x4y0" */; // synthesis attribute RLOC of ff_b3 "X4Y0"	

FDC ff_c0(.D(data),  .C(notclk), .CLR(rst), .Q(cz[0]))
/*synthesis rloc = "x1y1" iob = "FALSE" */ ; 
// synthesis attribute RLOC of ff_c0 "X1Y1"
// synthesis attribute IOB of ff_c0 "FALSE"
FDC ff_c1(.D(cz[0]), .C(clk90),  .CLR(rst), .Q(cz[1]))
/*synthesis rloc = "x2y1" */; // synthesis attribute RLOC of ff_c1 "X2Y1"
FDC ff_c2(.D(cz[1]), .C(clk),    .CLR(rst), .Q(cz[2]))
/*synthesis rloc = "x4y0" */; // synthesis attribute RLOC of ff_c2 "X4Y0"
FDC ff_c3(.D(cz[2]), .C(clk),    .CLR(rst), .Q(cz[3]))
/*synthesis rloc = "x4y1" */; // synthesis attribute RLOC of ff_c3 "X4Y1"	

FDC ff_d0(.D(data),  .C(notclk90), .CLR(rst), .Q(dz[0]))
/*synthesis rloc = "x0y1" iob = "FALSE" */ ; 
// synthesis attribute RLOC of ff_d0 "X0Y1"
// synthesis attribute IOB of ff_d0 "FALSE"
FDC ff_d1(.D(dz[0]), .C(notclk),   .CLR(rst), .Q(dz[1]))
/*synthesis rloc = "x1y1" */; // synthesis attribute RLOC of ff_d1 "X1Y1"
FDC ff_d2(.D(dz[1]), .C(clk90),    .CLR(rst), .Q(dz[2]))
/*synthesis rloc = "x2y1" */; // synthesis attribute RLOC of ff_d2 "X2Y1"
FDC ff_d3(.D(dz[2]), .C(clk),      .CLR(rst), .Q(dz[3]))
/*synthesis rloc = "x4y1" */; // synthesis attribute RLOC of ff_d3 "X4Y1"	

endmodule
