//////////////////////////////////////////////////////////////////////////////
//
//  Xilinx, Inc. 2002                 www.xilinx.com
//
//  XAPP 224 - Data Recovery
//
//////////////////////////////////////////////////////////////////////////////
//
//  File name :       data_recovery_virtexe.v
//
//  Description :     VirtexE data recovery module, uses two clocks 90 degrees apart to recover data
// 		      from incoming line data. (which is running slightly faster or slower than these
// 		      two clocks).
//
// 		      Output data bit 0 (sdata(0)) is valid when dv(0) is high
// 		      Output data bit 1 (sdata(1)) is valid when dv(1) is high
//
// 		      sample A is captured on : rising clk,    rising clk,   rising clk
// 		      sample B is captured on : rising clk90,  rising clk,   rising clk
// 		      sample C is captured on : falling clk,   rising clk,   rising clk
// 		      sample D is captured on : falling clk90, falling clk,  rising clk
//
//  Date - revision : May 16th 2003 - v 2.2
//
//  Author :          NJS
//
//  Contact : e-mail  hotline@xilinx.com
//            phone   + 1 800 255 7778 
//
//  Disclaimer: LIMITED WARRANTY AND DISCLAMER. These designs are
//              provided to you "as is". Xilinx and its licensors make and you
//              receive no warranties or conditions, express, implied,
//              statutory or otherwise, and Xilinx specifically disclaims any
//              implied warranties of merchantability, non-infringement,or
//              fitness for a particular purpose. Xilinx does not warrant that
//              the functions contained in these designs will meet your
//              requirements, or that the operation of these designs will be
//              uninterrupted or error free, or that defects in the Designs
//              will be corrected. Furthermore, Xilinx does not warrantor
//              make any representations regarding use or the results of the
//              use of the designs in terms of correctness, accuracy,
//              reliability, or otherwise.
//
//              LIMITATION OF LIABILITY. In no event will Xilinx or its
//              licensors be liable for any loss of data, lost profits,cost
//              or procurement of substitute goods or services, or for any
//              special, incidental, consequential, or indirect damages
//              arising from the use or operation of the designs or
//              accompanying documentation, however caused and on any theory
//              of liability. This limitation will apply even if Xilinx
//              has been advised of the possibility of such damage. This
//              limitation shall apply not-withstanding the failure of the
//              essential purpose of any limited remedies herein.
//
//  Copyright  2002 Xilinx, Inc.
//  All rights reserved
//
//////////////////////////////////////////////////////////////////////////////
//
`timescale 1 ps / 1ps

module data_recovery_virtexe (
	clk,
	clk90,
	data,
	rst,
	sdata,
	dv); 

input 		clk, clk90, data, rst ;
	
output 	[1 :0] 	sdata, dv ;
reg 	[1:0] 	sdata ;

wire 		notclk ;
wire 		notclk90;
wire 	[2:0] 	az ;
wire 	[2:0] 	bz ;
wire 	[2:0] 	cz ;
wire 	[2:0] 	dz ;
reg 		az3 ;
reg 		bz3 ;
reg 		cz3 ;
reg 		dz3 ;
reg 		aap, aan ;
reg 		bbp, bbn ;
reg 		ccp, ccn ;
reg 		ddp, ddn ;
reg 		usea;
reg 		useb;
reg 		usec;
reg 		used;
reg 		useaaa;
reg 		usebbb;
reg 		useccc;
reg 		useddd;
wire 	[1:0] 	sdataa;
wire 	[1:0] 	sdatab ;
wire 	[1:0] 	sdatac;
wire 	[1:0] 	sdatad;
reg 	[1:0] 	dvalid;
reg 		pipe_ce0;

assign notclk = ~clk ;
assign notclk90 = ~clk90 ;
assign sdataa[0] = az3 & useaaa ; assign sdataa[1] = az3 & useddd ;
assign sdatab[0] = bz3 & usebbb ; assign sdatab[1] = 1'b0 ;
assign sdatac[0] = cz3 & useccc ; assign sdatac[1] = 1'b0 ;
assign sdatad[0] = dz3 & useddd ; assign sdatad[1] = dz3 & useaaa ;
assign dv = dvalid ;

always @(posedge clk or posedge rst)
begin
if (rst)
  begin
	dvalid <= 2'b00 ;
	az3 <= 1'b1 ; bz3 <= 1'b1 ; cz3 <= 1'b1 ; dz3 <= 1'b1 ;
	aan <= 1'b0  ; bbn <= 1'b0 ;  ccn <= 1'b0 ;  ddn <= 1'b0 ;	
	aap <= 1'b0  ; bbp <= 1'b0 ;  ccp <= 1'b0 ;  ddp <= 1'b0 ;	
	usea <= 1'b0  ; useb <= 1'b0  ; usec <= 1'b0  ; used <= 1'b0 ;
	useaaa <= 1'b0 ; usebbb <= 1'b0 ; useccc <= 1'b0 ; useddd <= 1'b0 ;
	sdata <= 2'b11 ; pipe_ce0 <= 1'b0 ;
  end // if
else begin
	az3 <= az[2] ;
	bz3 <= bz[2] ;
	cz3 <= cz[2] ;
	dz3 <= dz[2] ;
	aap <= (az[1] ^ az[2]) & ~az[1] ;	// find positive edges
	bbp <= (bz[1] ^ bz[2]) & ~bz[1] ;
	ccp <= (cz[1] ^ cz[2]) & ~cz[1] ;
	ddp <= (dz[1] ^ dz[2]) & ~dz[1] ;
	aan <= (az[1] ^ az[2]) & az[1] ;	// find negative edges
	bbn <= (bz[1] ^ bz[2]) & bz[1] ;
	ccn <= (cz[1] ^ cz[2]) & cz[1] ;
	ddn <= (dz[1] ^ dz[2]) & dz[1] ;
	usea <= (bbp & ~ccp & ~ddp & aap) | (bbn & ~ccn & ~ddn & aan) ;
	useb <= (ccp & ~ddp & aap & bbp) | (ccn & ~ddn & aan & bbn) ;
	usec <= (ddp & aap & bbp & ccp) | (ddn & aan & bbn & ccn) ;
	used <= (aap & ~bbp & ~ccp & ~ddp) | (aan & ~bbn & ~ccn & ~ddn) ;
	if (usea | useb | usec | used) begin
	   	pipe_ce0 <= 1'b1 ;
		useaaa <= usea ;
		usebbb <= useb ;
		useccc <= usec ;
		useddd <= used ;
	end
	if (pipe_ce0) begin
		sdata <= sdataa | sdatab | sdatac | sdatad ;
	end
	if (useddd & usea) begin
		dvalid <= 2'b00 ;
	end else if (useaaa & used) begin
		dvalid <= 2'b11 ; 
	end else begin
		dvalid <= 2'b01 ;
	end
  end // if
end //always

// get all the samples into the same time domain 

FDC ff_a0(.D(data),  .C(clk), .CLR(rst), .Q(az[0]))
/*synthesis rloc = "R0C0.S0" */; // synthesis attribute RLOC of ff_a0 "R0C0.S0"
FDC ff_a1(.D(az[0]), .C(clk), .CLR(rst), .Q(az[1]))
/*synthesis rloc = "R0C0.S0" */; // synthesis attribute RLOC of ff_a1 "R0C0.S0"
FDC ff_a2(.D(az[1]), .C(clk), .CLR(rst), .Q(az[2]))
/*synthesis rloc = "R0C3.S1" */; // synthesis attribute RLOC of ff_a2 "R0C3.S1"

FDC ff_b0(.D(data),  .C(clk90), .CLR(rst), .Q(bz[0]))
/*synthesis rloc = "R0C1.S0" */; // synthesis attribute RLOC of ff_b0 "R0C1.S0"
FDC ff_b1(.D(bz[0]), .C(clk),   .CLR(rst), .Q(bz[1]))
/*synthesis rloc = "R0C3.S0" */; // synthesis attribute RLOC of ff_b1 "R0C3.S0"
FDC ff_b2(.D(bz[1]), .C(clk),   .CLR(rst), .Q(bz[2]))
/*synthesis rloc = "R0C3.S0" */; // synthesis attribute RLOC of ff_b2 "R0C3.S0"

FDC ff_c0(.D(data),  .C(notclk), .CLR(rst), .Q(cz[0]))
/*synthesis rloc = "R0C1.S1" */; // synthesis attribute RLOC of ff_c0 "R0C1.S1"
FDC ff_c1(.D(cz[0]), .C(clk90),  .CLR(rst), .Q(cz[1]))
/*synthesis rloc = "R0C2.S1" */; // synthesis attribute RLOC of ff_c1 "R0C2.S1"
FDC ff_c2(.D(cz[1]), .C(clk),    .CLR(rst), .Q(cz[2]))
/*synthesis rloc = "R0C3.S1" */; // synthesis attribute RLOC of ff_c2 "R0C3.S1"

FDC ff_d0(.D(data),  .C(notclk90), .CLR(rst), .Q(dz[0]))
/*synthesis rloc = "R0C0.S1" */; // synthesis attribute RLOC of ff_d0 "R0C0.S1"
FDC ff_d1(.D(dz[0]), .C(notclk),   .CLR(rst), .Q(dz[1]))
/*synthesis rloc = "R0C1.S1" */; // synthesis attribute RLOC of ff_d1 "R0C1.S1"
FDC ff_d2(.D(dz[1]), .C(clk90),    .CLR(rst), .Q(dz[2]))
/*synthesis rloc = "R0C2.S0" */; // synthesis attribute RLOC of ff_d2 "R0C2.S0"

endmodule
